@@ -12,7 +12,7 @@ from debian_linux import config
 from debian_linux.debian import PackageDescription, PackageRelation, \
     PackageRelationEntry, PackageRelationGroup, VersionLinux
 from debian_linux.gencontrol import Gencontrol as Base, merge_packages, \
-    iter_featuresets
+    iter_featuresets, iter_flavours
 from debian_linux.utils import Templates, read_control
 
 locale.setlocale(locale.LC_CTYPE, "C.UTF-8")
@@ -49,9 +49,9 @@ class Gencontrol(Base):
         },
         'packages': {
             'docs': config.SchemaItemBoolean(),
-            'headers-all': config.SchemaItemBoolean(),
             'installer': config.SchemaItemBoolean(),
             'libc-dev': config.SchemaItemBoolean(),
+            'meta': config.SchemaItemBoolean(),
             'tools-unversioned': config.SchemaItemBoolean(),
             'tools-versioned': config.SchemaItemBoolean(),
             'source': config.SchemaItemBoolean(),
@@ -91,10 +91,6 @@ class Gencontrol(Base):
             if src in data or not optional:
                 makeflags[dst] = data[src]
 
-    def _substitute_file(self, template, vars, target, append=False):
-        with open(target, 'a' if append else 'w') as f:
-            f.write(self.substitute(self.templates[template], vars))
-
     def do_main_setup(self, vars, makeflags, extra):
         super(Gencontrol, self).do_main_setup(vars, makeflags, extra)
         makeflags.update({
@@ -103,7 +99,8 @@ class Gencontrol(Base):
             'ABINAME': self.abiname_version + self.abiname_part,
             'SOURCEVERSION': self.version.complete,
         })
-        makeflags['SOURCE_BASENAME'] = self.vars['source_basename']
+        makeflags['SOURCE_BASENAME'] = vars['source_basename']
+        makeflags['SOURCE_SUFFIX'] = vars['source_suffix']
 
         # Prepare to generate debian/tests/control
         self.tests_control = self.process_packages(
@@ -183,23 +180,44 @@ class Gencontrol(Base):
 
     def do_main_packages(self, packages, vars, makeflags, extra):
         packages.extend(self.process_packages(
-            self.templates["control.main"], self.vars))
+            self.templates["control.main"], vars))
+
+        # Only build the metapackages if their names won't exactly match
+        # the packages they depend on
+        do_meta = self.config.merge('packages').get('meta', True) \
+            and vars['source_suffix'] != '-' + vars['version']
+
         if self.config.merge('packages').get('docs', True):
             packages.extend(self.process_packages(
-                self.templates["control.docs"], self.vars))
+                self.templates["control.docs"], vars))
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.docs.meta"], vars))
+                self.substitute_debhelper_config(
+                    'docs.meta', vars,
+                    '%(source_basename)s-doc%(source_suffix)s' % vars)
         if self.config.merge('packages').get('tools-unversioned', True):
             packages.extend(self.process_packages(
-                self.templates["control.tools-unversioned"], self.vars))
+                self.templates["control.tools-unversioned"], vars))
         if self.config.merge('packages').get('tools-versioned', True):
             packages.extend(self.process_packages(
-                self.templates["control.tools-versioned"], self.vars))
+                self.templates["control.tools-versioned"], vars))
+            self.substitute_debhelper_config('perf', vars,
+                                              'linux-perf-%(version)s' % vars)
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.tools-versioned.meta"], vars))
+                self.substitute_debhelper_config('perf.meta', vars,
+                                                  'linux-perf')
         if self.config.merge('packages').get('source', True):
             packages.extend(self.process_packages(
-                self.templates["control.sourcebin"], self.vars))
-
-        self._substitute_file('perf.lintian-overrides', self.vars,
-                              'debian/linux-perf-%s.lintian-overrides' %
-                              self.vars['version'])
+                self.templates["control.sourcebin"], vars))
+            if do_meta:
+                packages.extend(self.process_packages(
+                    self.templates["control.sourcebin.meta"], vars))
+                self.substitute_debhelper_config(
+                    'sourcebin.meta', vars,
+                    '%(source_basename)s-source%(source_suffix)s' % vars)
 
     def do_indep_featureset_setup(self, vars, makeflags, featureset, extra):
         makeflags['LOCALVERSION'] = vars['localversion']
@@ -265,25 +283,9 @@ class Gencontrol(Base):
         else:
             build_signed = False
 
-        # Some userland architectures require kernels from another
-        # (Debian) architecture, e.g. x32/amd64.
-        # And some derivatives don't need the headers-all packages
-        # for other reasons.
-        if self.config['base', arch].get('featuresets') and \
-           self.config.merge('packages').get('headers-all', True):
-            headers_arch = self.templates["control.headers.arch"]
-            packages_headers_arch = self.process_packages(headers_arch, vars)
-            packages_headers_arch[-1]['Depends'].extend(PackageRelation())
-            extra['headers_arch_depends'] = (
-                packages_headers_arch[-1]['Depends'])
-        else:
-            packages_headers_arch = []
-
         if self.config.merge('packages').get('libc-dev', True):
             libc_dev = self.templates["control.libc-dev"]
-            packages_headers_arch[0:0] = self.process_packages(libc_dev, {})
-
-        merge_packages(packages, packages_headers_arch, arch)
+            merge_packages(packages, self.process_packages(libc_dev, {}), arch)
 
         if self.config['base', arch].get('featuresets') and \
            self.config.merge('packages').get('source', True):
@@ -334,12 +336,25 @@ class Gencontrol(Base):
         vars['localversion_headers'] = vars['localversion']
         makeflags['LOCALVERSION_HEADERS'] = vars['localversion_headers']
 
+        self.default_flavour = self.config.merge('base', arch, featureset) \
+                                          .get('default-flavour')
+        if self.default_flavour is not None:
+            if featureset != 'none':
+                raise RuntimeError("default-flavour set for %s %s,"
+                                   " but must only be set for featureset none"
+                                   % (arch, featureset))
+            if self.default_flavour \
+               not in iter_flavours(self.config, arch, featureset):
+                raise RuntimeError("default-flavour %s for %s %s does not exist"
+                                   % (self.default_flavour, arch, featureset))
+
     flavour_makeflags_base = (
         ('compiler', 'COMPILER', False),
         ('compiler-filename', 'COMPILER', True),
         ('kernel-arch', 'KERNEL_ARCH', False),
-        ('cflags', 'CFLAGS_KERNEL', True),
+        ('cflags', 'KCFLAGS', True),
         ('override-host-type', 'OVERRIDE_HOST_TYPE', True),
+        ('cross-compile-compat', 'CROSS_COMPILE_COMPAT', True),
     )
 
     flavour_makeflags_build = (
@@ -363,6 +378,7 @@ class Gencontrol(Base):
                                                flavour)
         config_image = self.config.merge('image', arch, featureset, flavour)
 
+        vars['flavour'] = vars['localversion'][1:]
         vars['class'] = config_description['hardware']
         vars['longclass'] = (config_description.get('hardware-long')
                              or vars['class'])
@@ -387,6 +403,7 @@ class Gencontrol(Base):
         headers = self.templates["control.headers"]
         assert len(headers) == 1
 
+        do_meta = self.config.merge('packages').get('meta', True)
         config_entry_base = self.config.merge('base', arch, featureset,
                                               flavour)
         config_entry_build = self.config.merge('build', arch, featureset,
@@ -495,6 +512,36 @@ class Gencontrol(Base):
             extra['headers_arch_depends'].append('%s (= ${binary:Version})' %
                                                  packages_own[-1]['Package'])
 
+        # The image meta-packages will depend on signed linux-image
+        # packages where applicable, so should be built from the
+        # signed source packages The header meta-packages will also be
+        # built along with the signed packages, to create a dependency
+        # relationship that ensures src:linux and src:linux-signed-*
+        # transition to testing together.
+        if do_meta and not build_signed:
+            packages_meta = self.process_packages(
+                self.templates['control.image.meta'], vars)
+            assert len(packages_meta) == 1
+            packages_meta += self.process_packages(
+                self.templates['control.headers.meta'], vars)
+            assert len(packages_meta) == 2
+
+            if flavour == self.default_flavour \
+               and not self.vars['source_suffix']:
+                packages_meta[0].setdefault('Provides', PackageRelation()) \
+                                .append('linux-image-generic')
+                packages_meta[1].setdefault('Provides', PackageRelation()) \
+                                .append('linux-headers-generic')
+
+            packages_own.extend(packages_meta)
+
+            self.substitute_debhelper_config(
+                "image.meta", vars,
+                "linux-image%(localversion)s" % vars)
+            self.substitute_debhelper_config(
+                'headers.meta', vars,
+                'linux-headers%(localversion)s' % vars)
+
         if config_entry_build.get('vdso', False):
             makeflags['VDSO'] = True
 
@@ -507,6 +554,12 @@ class Gencontrol(Base):
             makeflags['DEBUG'] = True
             packages_own.extend(self.process_packages(
                 self.templates['control.image-dbg'], vars))
+            if do_meta:
+                packages_own.extend(self.process_packages(
+                    self.templates["control.image-dbg.meta"], vars))
+                self.substitute_debhelper_config(
+                    'image-dbg.meta', vars,
+                    'linux-image%(localversion)s-dbg' % vars)
 
         merge_packages(packages, packages_own, arch)
 
@@ -577,9 +630,8 @@ class Gencontrol(Base):
         if build_signed:
             makeflags['KCONFIG_OPTIONS'] += ' -o MODULE_SIG=y'
         # Add "salt" to fix #872263
-        makeflags['KCONFIG_OPTIONS'] += (' -o "BUILD_SALT=\\"%s%s\\""' %
-                                         (vars['abiname'],
-                                          vars['localversion']))
+        makeflags['KCONFIG_OPTIONS'] += \
+            ' -o "BUILD_SALT=\\"%(abiname)s%(localversion)s\\""' % vars
 
         cmds_binary_arch = ["$(MAKE) -f debian/rules.real binary-arch-flavour "
                             "%s" %
@@ -601,22 +653,14 @@ class Gencontrol(Base):
                      cmds=["$(MAKE) -f debian/rules.real %s %s" %
                            (merged_config, makeflags)])
 
-        # Substitute kernel version etc. into maintainer scripts,
-        # translations and lintian overrides
-        self._substitute_file('headers.postinst', vars,
-                              'debian/linux-headers-%s%s.postinst' %
-                              (vars['abiname'], vars['localversion']))
-        for name in ['postinst', 'postrm', 'preinst', 'prerm']:
-            self._substitute_file('image.%s' % name, vars,
-                                  'debian/%s.%s' %
-                                  (image_main['Package'], name))
+        self.substitute_debhelper_config(
+            'headers', vars,
+            'linux-headers-%(abiname)s%(localversion)s' % vars)
+        self.substitute_debhelper_config('image', vars, image_main['Package'])
         if build_debug:
-            debug_lintian_over = (
-                'debian/linux-image-%s%s-dbg.lintian-overrides' %
-                (vars['abiname'], vars['localversion']))
-            self._substitute_file('image-dbg.lintian-overrides', vars,
-                                  debug_lintian_over)
-            os.chmod(debug_lintian_over, 0o755)
+            self.substitute_debhelper_config(
+                'image-dbg', vars,
+                'linux-image-%(abiname)s%(localversion)s-dbg' % vars)
 
     def process_changelog(self):
         version = self.version = self.changelog[0].version
@@ -637,6 +681,8 @@ class Gencontrol(Base):
             'source_package': self.changelog[0].source,
             'abiname': self.abiname_version + self.abiname_part,
         }
+        self.vars['source_suffix'] = \
+            self.changelog[0].source[len(self.vars['source_basename']):]
         self.config['version', ] = {'source': self.version.complete,
                                     'upstream': self.version.linux_upstream,
                                     'abiname_base': self.abiname_version,
